{
 "cells": [
  {
   "cell_type": "markdown",
   "metadata": {
    "collapsed": false
   },
   "source": [
    "## CS166: Elevator Simulation\n",
    "Group members: Galena, Urmila, Long, Valyo, Qiqi \n"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 42,
   "metadata": {
    "collapsed": false
   },
   "outputs": [
   ],
   "source": [
    "import random\n",
    "import numpy as np\n",
    "\n",
    "class Elevator():\n",
    "\n",
    "    \"\"\" Models an elevator \n",
    "    Attributes:\n",
    "        current_floor = An integer representing the elevator's current floor\n",
    "        num_floors = An integer representing the number of floors the elevator can travel\n",
    "        action_queue = A list representing passengers that need to be delivered to their destination\n",
    "        max_capacity = An integer indicating how many passengers can fit in the elevator \n",
    "        passengers_inside = A list of passenger objects whose pickedup attribute == True. That is, a list of passengers inside the elvator. \"\"\"\n",
    "\n",
    "    def __init__(self, num_floors, current_floor = 1, direction=\"up\", max_capacity = 300): #user must manually update max_capacity to change the elevator capacity.\n",
    "\n",
    "        self.current_floor = current_floor\n",
    "        self.num_floors = num_floors\n",
    "        self.action_queue = []  # next floor to stop at\n",
    "        self.max_capacity = max_capacity \n",
    "        self.passengers_inside = [] # list of passengers inside the elevator\n",
    "\n",
    "    def add_request(self, passenger):\n",
    "\n",
    "        \"\"\" Add passengers to the action queue if capacity allows \"\"\"\n",
    "        if len(self.passengers_inside) <= self.max_capacity :\n",
    "            self.passengers_inside.append(passenger)  \n",
    "            self.action_queue.append(passenger.destination)\n",
    "            passenger.pickedup = True\n",
    "\n",
    "    def completed(self, passenger, sim_time):\n",
    "\n",
    "        \"\"\" Remove passengers who have been delivered to their destination \"\"\"\n",
    "\n",
    "        self.passengers_inside.remove(passenger)\n",
    "        self.action_queue.remove(passenger.destination)\n",
    "        passenger.pickedup = False\n",
    "        passenger.reached = True\n",
    "        passenger.delivery_time = sim_time\n",
    "        print \"Next Actions:\",self.action_queue\n",
    "        \n",
    "    def move(self, curr_fl):\n",
    "        self.current_floor = curr_fl\n",
    "\n",
    "class Passenger():\n",
    "\n",
    "    \"\"\" Models a passenger\n",
    "    Attributes:\n",
    "        starting_floor = An integer representing the floor the passenger starts on \n",
    "        destination = An integer representing the floor the passenger is going to \n",
    "        pickedup.reached= Both Boolean values indicating whether a paasenger is picked up / has reached destination \"\"\"\n",
    "\n",
    "    def __init__(self, num_floors, pickedup=False, reached=False):\n",
    "        self.starting_floor = random.randint(1, num_floors)\n",
    "        self.destination = random.choice([i for i in range(1,num_floors+1) if i!=self.starting_floor])\n",
    "        self.pickedup = pickedup #default: False\n",
    "        self.reached = reached #default: False\n",
    "        self.direction = \"up\" if self.destination > self.starting_floor else \"down\"\n",
    "        self.delivery_time = 0\n",
    "\n",
    "class Building():\n",
    "\n",
    "    \"\"\" Models a building\n",
    "    Attributes:\n",
    "        num_floors = An integer representing the number of floors in the building\n",
    "        num_passengers = An integer indicating how many passengers there are in the building\n",
    "        \"\"\"\n",
    "    \n",
    "\n",
    "    def __init__(self, num_passengers, num_floors=10):\n",
    "\n",
    "        self.num_floors = num_floors\n",
    "        self.num_passengers = num_passengers\n",
    "        self.passengers_remained = num_passengers\n",
    "        self.simulation_time = 0\n",
    "\n",
    "    def execution_simple(self):\n",
    "        \"\"\" Implement the naive execution strategy:\n",
    "        The Elevator stops at every floor, picking up passengers, if any, regardless of their intended direction,\n",
    "        until all passengers have arrived at their respective destination\"\"\"\n",
    "        elevator = Elevator(self.num_floors)\n",
    "\n",
    "        pass_list = []\n",
    "        #to store inidividual passenger wait times\n",
    "        deliverytime_list = []\n",
    "        self.passenger_delivery_times=[]\n",
    "        \n",
    "        for x in range(self.num_passengers):\n",
    "            y = Passenger(self.num_floors)\n",
    "            pass_list.append(y)\n",
    "        \n",
    "        while self.passengers_remained > 0: #start loop, the elevator would keep going all the way up and all the way down until no passengers are left\n",
    "            for fl in range(0, self.num_floors):\n",
    "\n",
    "                elevator.move(fl+1) #move 1 floor and door opens, 12 seconds\n",
    "                self.simulation_time += 12\n",
    "                \n",
    "                for passenger in pass_list:\n",
    "\n",
    "                    if elevator.current_floor == passenger.starting_floor and passenger.reached==False and passenger.pickedup==False:\n",
    "\n",
    "                        elevator.add_request(passenger) \n",
    "                        self.simulation_time += 10\n",
    "\n",
    "                    if elevator.current_floor == passenger.destination and passenger.reached==False and passenger.pickedup==True:\n",
    "                        elevator.completed(passenger, self.simulation_time)\n",
    "                        self.passengers_remained -= 1\n",
    "                        deliverytime_list.append(self.simulation_time)\n",
    "                        print \"Total number of remaining passengers:\",(self.passengers_remained)\n",
    "                        if self.passengers_remained == 0:                        \n",
    "                            self.passenger_delivery_times=deliverytime_list\n",
    "                            print \"Simulation Complete; no passengers remaining\"\n",
    "                            print \"Total time take:\", self.simulation_time\n",
    "                            return \n",
    "\n",
    "\n",
    "            for fl in range(self.num_floors+1, 1, -1):\n",
    "                \n",
    "                elevator.move(fl-1) #move 1 floor and door opens, 12 seconds\n",
    "                self.simulation_time += 12\n",
    "                \n",
    "                for passenger in pass_list:\n",
    "\n",
    "                    if elevator.current_floor == passenger.starting_floor and passenger.reached==False and passenger.pickedup==False:\n",
    "                        elevator.add_request(passenger) \n",
    "                        \n",
    "                    if elevator.current_floor == passenger.destination and passenger.reached==False and passenger.pickedup==True:\n",
    "                        elevator.completed(passenger, self.simulation_time) \n",
    "                        self.passengers_remained -= 1\n",
    "                        deliverytime_list.append(self.simulation_time)\n",
    "\n",
    "                        print \"Total number of remaining passengers:\", self.passengers_remained\n",
    "                        if self.passengers_remained == 0:\n",
    "                            self.passenger_delivery_times=deliverytime_list\n",
    "                            print \"Simulation Complete; no passengers remaining\"\n",
    "                            print \"Total time take:\",self.simulation_time\n",
    "                            return\n",
    "                        \n",
    "\n",
    "                    \n",
    "    def execution(self):\n",
    "        \"\"\" Implement the improved strategy\n",
    "        The elevator only picks up passengers that are going in the same direction as it is moving towards. \n",
    "        The iteration ends when all passengers have arrived at their respective destination\"\"\"\n",
    "        #initiate pass list\n",
    "        elevator = Elevator(self.num_floors)\n",
    "\n",
    "        pass_list = []\n",
    "        #to store inidividual passenger wait times\n",
    "        deliverytime_list = []\n",
    "        self.passenger_delivery_times=[]\n",
    "        \n",
    "        for x in range(self.num_passengers):\n",
    "            y = Passenger(self.num_floors)\n",
    "            pass_list.append(y)\n",
    "            \n",
    "        while self.passengers_remained > 0: #start loop, the elevator would keep going all the way up and all the way down until no passengers are left\n",
    "            for fl in range(0, self.num_floors):\n",
    "                door_open = False\n",
    "                elevator.move(fl+1) #move floors, 2 seconds\n",
    "                self.simulation_time += 2\n",
    "                \n",
    "                for passenger in pass_list:\n",
    "                    \n",
    "                    if elevator.current_floor == passenger.starting_floor and passenger.reached==False and passenger.pickedup==False and passenger.direction==\"up\":\n",
    "                        elevator.add_request(passenger)\n",
    "                        door_open = True\n",
    "\n",
    "                    if elevator.current_floor == passenger.destination and passenger.reached==False and passenger.pickedup==True:\n",
    "                        elevator.completed(passenger, self.simulation_time)\n",
    "                        self.passengers_remained -= 1\n",
    "                        deliverytime_list.append(self.simulation_time)\n",
    "                        \n",
    "                        print \"Total number of remaining passengers:\", self.passengers_remained\n",
    "                        door_open = True\n",
    "                        \n",
    "                    if self.passengers_remained == 0:\n",
    "                        self.passenger_delivery_times=deliverytime_list\n",
    "                        self.simulation_time =+ 10 #add 10 secs for door opening\n",
    "                        print \"Simulation Complete; no passengers remaining\"\n",
    "                        \n",
    "                        print \"Total time taken:\", self.simulation_time\n",
    "                        return\n",
    "                if door_open: #door opens, add 10 seconds\n",
    "                        self.simulation_time += 10\n",
    "\n",
    "            for fl in range(self.num_floors+1, 1, -1):\n",
    "                elevator.move(fl-1) #move floors, 2 seconds\n",
    "                self.simulation_time += 2\n",
    "                \n",
    "                for passenger in pass_list:\n",
    "                    door_open = False\n",
    "                    \n",
    "                    if elevator.current_floor == passenger.starting_floor and passenger.reached==False and passenger.pickedup==False and passenger.direction == \"down\":\n",
    "\n",
    "                        elevator.add_request(passenger)\n",
    "                        door_open = True\n",
    "\n",
    "                    if elevator.current_floor == passenger.destination and passenger.reached==False and passenger.pickedup==True:\n",
    "                        elevator.completed(passenger, self.simulation_time)               \n",
    "                        self.passengers_remained -= 1\n",
    "                        deliverytime_list.append(self.simulation_time)\n",
    "                        \n",
    "                        print \"Total number of remaining passengers:\", self.passengers_remained\n",
    "                        door_open = True\n",
    "                    if door_open :\n",
    "                        self.simulation_time += 10\n",
    "                    \n",
    "                    if self.passengers_remained == 0:                            \n",
    "                        self.passenger_delivery_times=deliverytime_list\n",
    "                        print \"Simulation Complete; no passengers remaining\"\n",
    "                        print \"Total time taken:\", self.simulation_time\n",
    "                        return"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 16,
   "metadata": {
    "collapsed": false
   },
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "Next Actions: [6, 9, 10]\n",
      "Total number of remaining passengers: 19\n",
      "Next Actions: [9, 10, 3]\n",
      "Total number of remaining passengers: 18\n",
      "Next Actions: [10, 3, 5]\n",
      "Total number of remaining passengers: 17\n",
      "Next Actions: [3, 5, 6]\n",
      "Total number of remaining passengers: 16\n",
      "Next Actions: [3, 5, 1]\n",
      "Total number of remaining passengers: 15\n",
      "Next Actions: [3, 1]\n",
      "Total number of remaining passengers: 14\n",
      "Next Actions: [1, 10, 9]\n",
      "Total number of remaining passengers: 13\n",
      "Next Actions: [10, 9, 8]\n",
      "Total number of remaining passengers: 12\n",
      "Next Actions: [10, 9, 8]\n",
      "Total number of remaining passengers: 11\n",
      "Next Actions: [10, 9, 10]\n",
      "Total number of remaining passengers: 10\n",
      "Next Actions: [10, 10, 3]\n",
      "Total number of remaining passengers: 9\n",
      "Next Actions: [10, 3, 3]\n",
      "Total number of remaining passengers: 8\n",
      "Next Actions: [3, 3]\n",
      "Total number of remaining passengers: 7\n",
      "Next Actions: [3, 3, 10]\n",
      "Total number of remaining passengers: 6\n",
      "Next Actions: [3, 10]\n",
      "Total number of remaining passengers: 5\n",
      "Next Actions: [10]\n",
      "Total number of remaining passengers: 4\n",
      "Next Actions: [2, 5, 2]\n",
      "Total number of remaining passengers: 3\n",
      "Next Actions: [2, 2]\n",
      "Total number of remaining passengers: 2\n",
      "Next Actions: [2]\n",
      "Total number of remaining passengers: 1\n",
      "Next Actions: []\n",
      "Total number of remaining passengers: 0\n",
      "Simulation Complete; no passengers remaining\n",
      "Total time take: 1028\n"
     ]
    }
   ],
   "source": [
    "#sanity check, one iteration\n",
    "build_simple = Building(20)\n",
    "build_simple.execution_simple()"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 17,
   "metadata": {
    "collapsed": false
   },
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "average wait time: 586.0\n",
      "maximum wait time: 1028\n"
     ]
    }
   ],
   "source": [
    "#sanity check: 1 iteration\n",
    "import numpy as np\n",
    "simple_wait_times = build_simple.passenger_delivery_times\n",
    "simple_mean = np.mean(simple_wait_times)\n",
    "simple_max = np.max(simple_wait_times)\n",
    "print \"average wait time:\",simple_mean\n",
    "print \"maximum wait time:\",simple_max"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 18,
   "metadata": {
    "collapsed": false
   },
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "Next Actions: [9, 10, 10]\n",
      "Total number of remaining passengers: 19\n",
      "Next Actions: [10, 10, 10]\n",
      "Total number of remaining passengers: 18\n",
      "Next Actions: [10, 10]\n",
      "Total number of remaining passengers: 17\n",
      "Next Actions: [10]\n",
      "Total number of remaining passengers: 16\n",
      "Next Actions: []\n",
      "Total number of remaining passengers: 15\n",
      "Next Actions: [2, 2, 8]\n",
      "Total number of remaining passengers: 14\n",
      "Next Actions: [2, 2]\n",
      "Total number of remaining passengers: 13\n",
      "Next Actions: [2, 2, 2]\n",
      "Total number of remaining passengers: 12\n",
      "Next Actions: [2, 2]\n",
      "Total number of remaining passengers: 11\n",
      "Next Actions: [2]\n",
      "Total number of remaining passengers: 10\n",
      "Next Actions: []\n",
      "Total number of remaining passengers: 9\n",
      "Next Actions: [10, 9]\n",
      "Total number of remaining passengers: 8\n",
      "Next Actions: [10]\n",
      "Total number of remaining passengers: 7\n",
      "Next Actions: []\n",
      "Total number of remaining passengers: 6\n",
      "Next Actions: [7, 4, 2]\n",
      "Total number of remaining passengers: 5\n",
      "Next Actions: [4, 2]\n",
      "Total number of remaining passengers: 4\n",
      "Next Actions: [2, 1, 1]\n",
      "Total number of remaining passengers: 3\n",
      "Next Actions: [1, 1]\n",
      "Total number of remaining passengers: 2\n",
      "Next Actions: [1]\n",
      "Total number of remaining passengers: 1\n",
      "Next Actions: []\n",
      "Total number of remaining passengers: 0\n",
      "Simulation Complete; no passengers remaining\n",
      "Total time taken: 500\n"
     ]
    }
   ],
   "source": [
    "#sanity check: 1 iteration\n",
    "build_improved = Building(20)\n",
    "build_improved.execution()"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 20,
   "metadata": {
    "collapsed": false
   },
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "average wait time: 274.4\n",
      "maximum wait time: 490\n"
     ]
    }
   ],
   "source": [
    "#sanity check: 1 iteration\n",
    "improved_wait_times = build_improved.passenger_delivery_times\n",
    "improved_mean = np.mean(improved_wait_times)\n",
    "improved_max = np.max(improved_wait_times)\n",
    "print \"average wait time:\",improved_mean\n",
    "print \"maximum wait time:\",improved_max"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 44,
   "metadata": {
    "collapsed": false
   },
   "outputs": [
   ],
   "source": [
    "%%capture\n",
    "#simulation: 1000 iterations\n",
    "simple_wait_times_mean = []\n",
    "simple_wait_times_max = []\n",
    "for i in range(1000):\n",
    "    build_simple = Building(20)\n",
    "    build_simple.execution_simple()\n",
    "    simple_wait_times_mean.append(np.mean(build_simple.passenger_delivery_times))\n",
    "    simple_wait_times_max.append(np.max(build_simple.passenger_delivery_times))"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 45,
   "metadata": {
    "collapsed": false
   },
   "outputs": [
   ],
   "source": [
    "%%capture\n",
    "#simulation: 1000 iterations\n",
    "improved_wait_times_mean = []\n",
    "improved_wait_times_max = []\n",
    "for i in range(1000):\n",
    "    build_improved = Building(20)\n",
    "    build_improved.execution()\n",
    "    improved_wait_times_mean.append(np.mean(build_improved.passenger_delivery_times))\n",
    "    improved_wait_times_max.append(np.max(build_improved.passenger_delivery_times))"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 46,
   "metadata": {
    "collapsed": false
   },
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "('Average wait time, naive strat, no cap limit, after 1000 realizations: ', 317.85)\n",
      "('Max wait time, naive strat, no cap limit, after 1000 realizations: ', 440.0)\n",
      "('Average wait time, improved strat, no cap limit, after 1000 realizations: ', 165.94)\n",
      "('Max wait time, improved strat, no cap limit, after 1000 realizations: ', 420.0)\n"
     ]
    }
   ],
   "source": [
    "#result print-out: case 1: no cap limit\n",
    "print(\"Average wait time, naive strat, no cap limit, after 1000 realizations: \", round(np.mean(simple_wait_times_mean), 2))\n",
    "print(\"Max wait time, naive strat, no cap limit, after 1000 realizations: \", round(np.max(simple_wait_times_max),2))\n",
    "print(\"Average wait time, improved strat, no cap limit, after 1000 realizations: \", round(np.mean(improved_wait_times_mean),2))\n",
    "print(\"Max wait time, improved strat, no cap limit, after 1000 realizations: \", round(np.max(improved_wait_times_max),2))"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 41,
   "metadata": {
    "collapsed": false
   },
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "('Average wait time, naive strat, with cap limit, after 1000 realizations: ', 522.36)\n",
      "('Max wait time, naive strat, with cap limit, after 1000 realizations: ', 1440.0)\n",
      "('Average wait time, improved strat, with cap limit, after 1000 realizations: ', 229.89)\n",
      "('Max wait time, improved strat, with cap limit, after 1000 realizations: ', 716.0)\n"
     ]
    }
   ],
   "source": [
    "#result print-out: case 2: with cap limit\n",
    "print(\"Average wait time, naive strat, with cap limit, after 1000 realizations: \", round(np.mean(simple_wait_times_mean),2))\n",
    "print(\"Max wait time, naive strat, with cap limit, after 1000 realizations: \", round(np.max(simple_wait_times_max),2))\n",
    "print(\"Average wait time, improved strat, with cap limit, after 1000 realizations: \", round(np.mean(improved_wait_times_mean),2))\n",
    "print(\"Max wait time, improved strat, with cap limit, after 1000 realizations: \", round(np.max(improved_wait_times_max),2))"
   ]
  }
 ],
 "metadata": {
  "anaconda-cloud": {
  },
  "kernelspec": {
   "display_name": "Python 2 (SageMath)",
   "language": "python",
   "name": "python2"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 2
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython2",
   "version": "2.7.14"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 0
}